@@ -11,6 +11,7 @@ import shortuuid
 import hashlib
 from json_logic import jsonLogic
 from enum import Enum
+from airflow.models import Variable
 
 DT_FMT_HMSf = '%H%M%S%f'
 
@@ -182,25 +183,13 @@ module_id_dict = {'Campaigns': '1', 'Invoice': '2', 'SalesOrder': '3', 'Purchase
                   'Predicting': '91', 'Notifications': '95'}
 
 
-def get_session_name(user_access_key, vtiger_url, vtiger_username):
-    # get challenge
-    headers = {'content-type': 'application/json'}
+def get_session_name(vtiger_access):
+    vtiger_url = vtiger_access.get('vtiger_url')
+    vtiger_username = vtiger_access.get('vtiger_username')
+    access_key_encoded = vtiger_access.get('vtiger_access_key')
 
-    challenge_url = '{vtiger_url}/webservice.php?operation=getchallenge&username={vtiger_username}'.format(
-        vtiger_url=vtiger_url, vtiger_username=vtiger_username)
+    if vtiger_url and vtiger_username and access_key_encoded:
 
-    response, status = invoke_http_request(challenge_url, 'GET', headers=headers)
-
-    if status == 200 and response.get('result', ''):
-        token = response.get('result').get('token')
-
-        # get md5 encoded access key
-
-        access_key = token + user_access_key
-        result = hashlib.md5(access_key.encode())
-        access_key_encoded = result.hexdigest()
-
-        # login
         login_url = "{vtiger_url}/webservice.php".format(vtiger_url=vtiger_url)
         payload = 'operation=login&username={vtiger_username}&accessKey={access_key}'.format(
             access_key=access_key_encoded, vtiger_username=vtiger_username)
@@ -209,9 +198,12 @@ def get_session_name(user_access_key, vtiger_url, vtiger_username):
         response, status = invoke_http_request(login_url, 'POST', headers, payload)
 
         if status == 200 and response.get('result', ''):
+            print("existing session expired. New session id is: ", response.get('result').get('sessionName'))
             return response.get('result').get('sessionName')
         else:
             print("Can not get Session name from Vtiger. Error occurred with Status", status)
+    else:
+        print("unable to get required data from vtiger_access")
 
 
 def trigger_workflow(workflow, event_type, data, service_url, api_access_key):
@@ -342,6 +334,21 @@ def run_external_workflow(conf, task_id, vtiger_access):
 
                 print(f"Response: {response}")
 
+                if response:
+                    if not response.get('success', ''):
+                        if response.get('error', ''):
+                            if response.get('error').get('code') in ["INVALID_USER_CREDENTIALS", "INVALID_SESSIONID"]:
+                                session_name = get_session_name(vtiger_access)
+
+                                # retry to execute query
+
+                                url = '{vtiger_url}/webservice.php?operation=query&sessionName={sessionName}&query={query}'.format(
+                                    sessionName=session_name, query=query, vtiger_url=vtiger_access.get('vtiger_url'))
+                                response, status = invoke_http_request(url, request_type, headers)
+
+                                vtiger_access['vtiger_session'] = session_name
+                                Variable.update(key='vtiger_access', value=vtiger_access, serialize_json=True)
+
                 if response:
                     # trigger only if condition is satisfied
                     if response.get('result') and workflow and event_type:
@@ -358,9 +365,11 @@ def run_external_workflow(conf, task_id, vtiger_access):
                 # trigger without condition
                 if event_type == 'import' or event_type == 'manual':
                     list_of_data = [data]
-                    trigger_workflow(workflow, event_type, list_of_data, vtiger_access.get('service_url'), vtiger_access.get('api_access_key'))
+                    trigger_workflow(workflow, event_type, list_of_data, vtiger_access.get('service_url'),
+                                     vtiger_access.get('api_access_key'))
                 else:
-                    trigger_workflow(workflow, event_type, data, vtiger_access.get('service_url'), vtiger_access.get('api_access_key'))
+                    trigger_workflow(workflow, event_type, data, vtiger_access.get('service_url'),
+                                     vtiger_access.get('api_access_key'))
         else:
             print("external_workflow_config not found for task id:", task_id)
     else:
@@ -420,6 +429,20 @@ def trigger_set_value_task(set_value_fields, conf, rule, session_name, vtiger_ac
 
         response, status = invoke_http_request(url, request_type, headers, payload=payload)
 
+        if response:
+            if not response.get('success', ''):
+                if response.get('error', ''):
+                    if response.get('error').get('code') in ["INVALID_USER_CREDENTIALS", "INVALID_SESSIONID"]:
+                        session_name = get_session_name(vtiger_access)
+
+                        # retry to execute query
+
+                        payload = {'operation': 'revise', 'sessionName': session_name, 'element': json.dumps(element)}
+                        response, status = invoke_http_request(url, request_type, headers, payload=payload)
+
+                        vtiger_access['vtiger_session'] = session_name
+                        Variable.update(key='vtiger_access', value=vtiger_access, serialize_json=True)
+
         print("Response: {}".format(response))
         if is_success_request(status):
             return response
@@ -513,6 +536,20 @@ def invoke_set_value_task(conf, task_id, vtiger_access):
                 headers = {'content-type': 'application/json'}
                 request_type = 'GET'
                 response, status = invoke_http_request(url, request_type, headers)
+                if response:
+                    if not response.get('success', ''):
+                        if response.get('error', ''):
+                            if response.get('error').get('code') in ["INVALID_USER_CREDENTIALS", "INVALID_SESSIONID"]:
+                                session_name = get_session_name(vtiger_access)
+
+                                # retry to execute query
+
+                                url = '{vtiger_url}/webservice.php?operation=query&sessionName={sessionName}&query={query}'.format(
+                                    sessionName=session_name, query=query, vtiger_url=vtiger_access.get('vtiger_url'))
+                                response, status = invoke_http_request(url, request_type, headers)
+
+                                vtiger_access['vtiger_session'] = session_name
+                                Variable.update(key='vtiger_access', value=vtiger_access, serialize_json=True)
 
                 print("Query Response: ")
 
@@ -700,6 +737,54 @@ def add_delay_task(conf, task_id, vtiger_access):
     endpoint = '{service_url}/api/v1/add-delayed-data'.format(service_url=vtiger_access.get('service_url'))
     headers = {'content_type': 'application/json', 'Authorization': vtiger_access.get('api_access_key')}
 
-
     response, status = invoke_http_request(endpoint, 'POST', headers, json_data=payload, timeout=61)
     print("response from add delayed task", response, "status", status)
+
+
+def get_crm_records(vtiger_access, conf):
+    """ This function will get record id and will fetch record data for that id from Vtiger.
+       NOTE: set return value in context["dag_run"].conf['data'] while calling this function.
+       params:
+       vtiger_access: vtiger access dictionary
+       conf: conf object"""
+
+    if 'data' in conf and 'id' in conf.get('data'):
+
+        # call vtiger api to get data
+        record_id = conf['data']['id']
+        source_module = conf['data']['module']
+        id = "{module}x{record_id}".format(module=module_id_dict[source_module], record_id=record_id)
+        session_name = vtiger_access.get('vtiger_session', '')
+        print("Session name: {}".format(session_name))
+        if not session_name:
+            print("unable to get session id from vtiger access. Please try again")
+            return None
+
+        url = "{vtiger_url}/webservice.php?operation=retrieve&id={record_id}&sessionName={sessionName}".format(
+            record_id=id, sessionName=session_name, vtiger_url=vtiger_access.get('vtiger_url'))
+        response, status = invoke_http_request(url, 'GET', {'content_type': 'application/json'})
+
+        # request session if existing one is expired
+
+        if response:
+            if not response.get('success', ''):
+                if response.get('error', ''):
+                    if response.get('error').get('code') in ["INVALID_USER_CREDENTIALS", "INVALID_SESSIONID"]:
+                        session_name = get_session_name(vtiger_access)
+
+                        # retry to execute query
+
+                        url = "{vtiger_url}/webservice.php?operation=retrieve&id={record_id}&sessionName={sessionName}".format(
+                            record_id=id, sessionName=session_name, vtiger_url=vtiger_access.get('vtiger_url'))
+                        response, status = invoke_http_request(url, 'GET', {'content_type': 'application/json'})
+
+                        vtiger_access['vtiger_session'] = session_name
+                        Variable.update(key='vtiger_access', value=vtiger_access, serialize_json=True)
+
+        if 'success' in response and response['success'] and 'result' in response:
+            return response['result']
+        else:
+            print("unable to get record details from Vtiger. Response: ", response.text)
+
+    else:
+        print("unable to get id & data from conf object")
